from Fortuna import *
import random  # Base Cases (randrange, randint and choice) for comparison only
from datetime import datetime
from math import floor


def distribution_timer(func: staticmethod, *args, call_sig="f(x)", max_distribution=20):
    num_cycles = 100000
    start_time = datetime.now()
    results = [func(*args) for _ in range(num_cycles)]
    end_time = datetime.now()
    total_time = (end_time - start_time).total_seconds()
    print(f"{call_sig} x {num_cycles}: {round(total_time * 1000.0, 2)} ms")
    unique_results = set(results)
    if len(unique_results) <= max_distribution:
        result_obj = {
            key: f"{round(results.count(key) / (num_cycles / 100), 2)}%" for key in sorted(list(unique_results))
        }
        for key, val in result_obj.items():
            print(f" {key}: {val}")
    print("")


def floor_dice(rolls, sides):
    """ Fast but poor distribution properties, can cause platform dependant bias. """
    total = rolls
    for _ in range(rolls):
        total += floor(random.random() * sides)
    return total


def randrange_dice(rolls, sides):
    """ Slow but good distribution properties. """
    total = rolls
    for _ in range(rolls):
        total += random.randrange(sides)
    return total


if __name__ == "__main__":

    t0 = datetime.now()
    print("Fortuna 0.15.2 Sample Distribution and Performance Test Suite\n")

    print("Random Numbers")
    print(f"{'-' * 73}\n")
    distribution_timer(random.randrange, 10, call_sig="Base Case:\nrandom.randrange(10)")
    distribution_timer(random_below, 10, call_sig="Fortuna.random_below(10)")
    distribution_timer(random.randint, 1, 10, call_sig="Base Case:\nrandom.randint(1, 10)")
    distribution_timer(random_range, 1, 10, call_sig="Fortuna.random_range(1, 10)")
    distribution_timer(d, 10, call_sig="Fortuna.d(10)")
    distribution_timer(dice, 1, 10, call_sig="Fortuna.dice(1, 10)")
    distribution_timer(plus_or_minus, 5, call_sig="Fortuna.plus_or_minus(5)")
    distribution_timer(plus_or_minus_linear, 5, call_sig="Fortuna.plus_or_minus_linear(5)")
    distribution_timer(plus_or_minus_curve, 5, call_sig="Fortuna.plus_or_minus_curve(5)")
    print("")

    print("Random Truth")
    print(f"{'-' * 73}\n")
    distribution_timer(percent_true, 25, call_sig="Fortuna.percent_true(25)")
    print("")

    print("Random Values from a Sequence")
    print(f"{'-' * 73}\n")
    some_list = ["Alpha", "Beta", "Delta", "Eta", "Gamma", "Kappa", "Zeta"]
    print(f"some_list = {some_list}\n")
    distribution_timer(random.choice, some_list, call_sig="Base Case:\nrandom.choice(some_list)")
    distribution_timer(random_value, some_list, call_sig="Fortuna.random_value(some_list)")
    monty = Mostly(some_list)
    print(f"monty = Mostly(some_list)\n")
    distribution_timer(monty.mostly_front, call_sig="monty.mostly_front()")
    distribution_timer(monty.mostly_middle, call_sig="monty.mostly_middle()")
    distribution_timer(monty.mostly_back, call_sig="monty.mostly_back()")
    distribution_timer(monty.mostly_first, call_sig="monty.mostly_first()")
    distribution_timer(monty.mostly_center, call_sig="monty.mostly_center()")
    distribution_timer(monty.mostly_last, call_sig="monty.mostly_last()")
    distribution_timer(monty, call_sig="monty()")
    truffle_shuffle = RandomCycle(some_list)
    print(f"truffle_shuffle = RandomCycle(some_list)\n")
    distribution_timer(truffle_shuffle, call_sig="truffle_shuffle()")
    print("")

    print("Random Values by Weighted Table")
    print(f"{'-' * 73}\n")
    cumulative_weighted_table = (
        (7, "Apple"),
        (11, "Banana"),
        (13, "Cherry"),
        (23, "Grape"),
        (26, "Lime"),
        (30, "Orange"),
    )
    cumulative_weighted_choice = CumulativeWeightedChoice(cumulative_weighted_table)
    distribution_timer(cumulative_weighted_choice, call_sig="cumulative_weighted_choice()")
    relative_weighted_table = (
        (7, "Apple"),
        (4, "Banana"),
        (2, "Cherry"),
        (10, "Grape"),
        (3, "Lime"),
        (4, "Orange"),
    )
    relative_weighted_choice = RelativeWeightedChoice(relative_weighted_table)
    distribution_timer(relative_weighted_choice, call_sig="relative_weighted_choice()")
    print("")

    r, s = 10, 10
    print(f"Multi Dice: {r}d{s}")
    print(f"{'-' * 73}\n")
    distribution_timer(randrange_dice, r, s, call_sig=f"Base Case:\nrandrange_dice({r}, {s})"),
    distribution_timer(floor_dice, r, s, call_sig=f"Base Case:\nfloor_dice({r}, {s})"),
    distribution_timer(dice, r, s, call_sig=f"Fortuna.dice({r}, {s})"),
    print("")

    print(f"{'-' * 73}")
    total_test_time = round((datetime.now() - t0).total_seconds(), 2)
    print(f"Total Test Time: {total_test_time} sec\n")



from Fortuna import weighted_choice, min_max, RandomCycle, zero_cool

from fortuna_extras import distribution_timer


class MultiCat:
    __qualname__ = "MultiCat"

    def __init__(self, obj):
        self.rand_cat = RandomCycle(obj.keys())
        self.obj = {key: RandomCycle(itm) for key, itm in obj.items()}

    def __call__(self, cat=None):
        if cat in self.obj.keys():
            return self.obj[cat]()
        else:
            return self.obj[self.rand_cat()]()


def magic_table_a():
    loot = (
        (50, "Potion of healing"),
        (60, f"Spell scroll (cantrip) {random_spell(spell_level=0)}"),
        (70, "Potion of climbing"),
        (90, f"Spell scroll (1st level) {random_spell(spell_level=1)}"),
        (94, f"Spell scroll (2nd level) {random_spell(spell_level=2)}"),
        (98, "Potion of greater healing"),
        (99, "Bag of holding"),
        (100, "Driftglobe")
    )
    return weighted_choice(loot)


spells = MultiCat({
    "cantrip": (
        "Acid Splash", "Blade Ward", "Chill Touch", "Dancing Lights", "Fire Bolt", "Friends", "Light",
        "Mage Hand", "Mending", "Message", "Minor Illusion", "Poison Spray", "Prestidigitation",
        "Ray of Frost", "Shocking Grasp", "True Strike"
    ),
    "level 1": (
        "Alarm", "Burning Hands", "Charm Person", "Chromatic Orb", "Color Spray", "Comprehend Languages",
        "Detect Magic", "Disguise Self", "Expeditious Retreat", "False Life", "Feather Fall", "Find Familiar",
        "Fog Cloud", "Grease", "Identify", "Illusory Script", "Jump", "Longstrider", "Mage Armor",
        "Magic Missile", "Protection from Evil and Good", "Ray of Sickness", "Shield", "Silent Image",
        "Sleep", "Tasha’s Hideous Laughter", "Tenser’s Floating Disk", "Thunderwave", "Unseen Servant",
        "Witch Bolt"
    ),
    "level 2": (
        "Alter Self", "Arcane Lock", "Blindness/Deafness", "Blur", "Cloud of Daggers", "Continual Flame",
        "Crown of Madness", "Darkness", "Darkvision", "Detect Thoughts", "Enlarge/Reduce", "Flaming Sphere",
        "Gentle Repose", "Gust of Wind", "Hold Person", "Invisibility", "Knock", "Levitate", "Locate Object",
        "Magic Mouth", "Magic Weapon", "Melf’s Acid Arrow", "Mirror Image", "Misty Step", "Nystul’s Magic Aura",
        "Phantasmal Force", "Ray of Enfeeblement", "Rope Trick", "Scorching Ray", "See Invisibility", "Shatter",
        "Spider Climb", "Suggestion", "Web"
    )
})


def random_spell(spell_level=None):
    level = spell_level if spell_level else zero_cool(2)
    return spells(level)


if __name__ == "__main__":
    distribution_timer(random_spell, "level 2")



# distutils: language = c++

cdef extern from "Fortuna.hpp":
    int _random_range "random_range"(int, int)
    int _random_below "random_below"(int)
    int _d "d"(int)
    int _dice "dice"(int, int)
    int _min_max "min_max"(int, int, int)
    int _percent_true "percent_true"(int)
    int _plus_or_minus "plus_or_minus"(int)
    int _plus_or_minus_linear "plus_or_minus_linear"(int)
    int _plus_or_minus_curve "plus_or_minus_curve"(int)
    int _zero_flat "zero_flat"(int)
    int _zero_cool "zero_cool"(int)
    int _zero_extreme "zero_extreme"(int)
    int _max_cool "max_cool"(int)
    int _max_extreme "max_extreme"(int)
    int _mostly_middle "mostly_middle"(int)
    int _mostly_center "mostly_center"(int)

cdef int INT_MAX = 1000000000
cdef int INT_MIN = -1000000000

def random_range(int lo, int hi):
    assert INT_MIN <= lo <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    assert INT_MIN <= hi <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _random_range(lo, hi)

def random_below(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _random_below(num)

def d(int sides):
    assert 1 <= sides <= INT_MAX, f"Input Error, out of range (1..{INT_MAX})."
    return _d(sides)

def dice(int rolls, int sides):
    max_rolls = 10000000
    assert rolls <= max_rolls, f"Input Error, out of range (1..{max_rolls})."
    assert 1 <= rolls * sides <= INT_MAX * 2, f"Input Error, product of input args (rolls * sides) out of range (1..{INT_MAX * 2})."
    return _dice(rolls, sides)

def min_max(int n, int lo, int hi):
    input_data = (n, lo, hi)
    for itm in input_data:
        assert INT_MIN <= itm <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _min_max(n, lo, hi)

def percent_true(int num):
    assert 0 <= num <= 100, f"Input Error, out of range (0..100)."
    return _percent_true(num) == 1

def plus_or_minus(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _plus_or_minus(num)

def plus_or_minus_linear(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _plus_or_minus_linear(num)

def plus_or_minus_curve(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _plus_or_minus_curve(num)

def zero_flat(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _zero_flat(num)

def zero_cool(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _zero_cool(num)

def zero_extreme(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _zero_extreme(num)

def max_cool(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _max_cool(num)

def max_extreme(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _max_extreme(num)

def mostly_middle(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _mostly_middle(num)

def mostly_center(int num):
    assert INT_MIN <= num <= INT_MAX, f"Input Error, out of range ({INT_MIN}..{INT_MAX})."
    return _mostly_center(num)

def random_value(arr):
    assert 1 <= len(arr) <= INT_MAX, f"Input Error, sequence length out of range (1..{INT_MAX})."
    return arr[random_below(len(arr))]

def pop_random_value(arr):
    assert 1 <= len(arr) <= INT_MAX, f"Input Error, sequence length out of range (1..{INT_MAX})."
    return arr.pop(random_below(len(arr)))

def weighted_choice(table):
    assert 1 <= len(table) <= INT_MAX, f"Input Error, sequence length out of range (1..{INT_MAX})."
    max_weight = table[-1][0]
    rand = _random_below(max_weight)
    for weight, value in table:
        if weight > rand:
            return value


class Mostly:
    """ The Quantum Monty """

    def __init__(self, arr):
        size = len(arr)
        assert 3 <= size <= INT_MAX, f"Input Error, sequence length out of range (3..{INT_MAX})."
        self.data = tuple(arr)
        self.max_id = size - 1
        self.methods = (
            self.mostly_first,
            self.mostly_front,
            self.mostly_middle,
            self.mostly_center,
            self.mostly_middle,
            self.mostly_back,
            self.mostly_last
        )

    def mostly_flat(self):
        return self.data[zero_flat(self.max_id)]

    def mostly_front(self):
        return self.data[zero_cool(self.max_id)]

    def mostly_back(self):
        return self.data[max_cool(self.max_id)]

    def mostly_middle(self):
        return self.data[mostly_middle(self.max_id)]

    def mostly_first(self):
        return self.data[zero_extreme(self.max_id)]

    def mostly_last(self):
        return self.data[max_extreme(self.max_id)]

    def mostly_center(self):
        return self.data[mostly_center(self.max_id)]

    def __call__(self):
        return random_value(self.methods)()


class RandomCycle:

    def __init__(self, arr):
        from random import shuffle
        assert 3 <= len(arr) <= INT_MAX, f"Input Error, sequence length out of range (3..{INT_MAX})."
        self.data = list(arr)
        shuffle(self.data)
        self.next = self.data.pop()
        self.size = len(self.data)

    def __call__(self):
        result = self.next
        self.next = self.data.pop(_max_extreme(self.size - 1))
        self.data.insert(_zero_extreme(self.size - 2), result)
        return result


class WeightedChoice:
    data = ()

    def __call__(self):
        max_weight = self.data[-1][0]
        rand = _random_below(max_weight)
        for weight, value in self.data:
            if weight > rand:
                return value

    @staticmethod
    def _setup(weights, non_unique=False):
        if non_unique:
            pass
        else:
            size = len(weights)
            warn_non_unique = (
                "Sanity Check!",
                "  Weighted Values should be unique, pass non_unique=True during instantiation to bypass this check.",
                "  As a result: non-unique values will have their probabilities logically accumulated.",
                "  Relative Weights are summed, Cumulative Weights are over-lapped, but the effect is the same.",
            )
            assert size == len(set(v for _, v in weights)), "\n".join(warn_non_unique)

    @staticmethod
    def _optimize(data):
        prev_weight = 0
        for w_pair in data:
            w_pair[0], prev_weight = w_pair[0] - prev_weight, w_pair[0]
        return sorted(data, key=lambda x: x[0], reverse=True)

    @staticmethod
    def _package(data):
        cum_weight = 0
        for w_pair in data:
            cum_weight += w_pair[0]
            w_pair[0] = cum_weight
        return tuple(tuple(itm) for itm in data)


class RelativeWeightedChoice(WeightedChoice):

    def __init__(self, weights, non_unique=False):
        size = len(weights)
        assert 1 <= size <= INT_MAX, f"Input Error, sequence length out of range (1..{INT_MAX})."
        self._setup(weights, non_unique)
        optimized_data = sorted([list(itm) for itm in weights], key=lambda x: x[0], reverse=True)
        self.data = self._package(optimized_data)


class CumulativeWeightedChoice(WeightedChoice):

    def __init__(self, weights, non_unique=False):
        size = len(weights)
        assert 1 <= size <= INT_MAX, f"Input Error, sequence length out of range (1..{INT_MAX})."
        assert size == len(set(w for w, _ in weights)), "Cumulative Weights must be unique, because math."
        self._setup(weights, non_unique)
        optimized_data = self._optimize(sorted([list(itm) for itm in weights], key=lambda x: x[0]))
        self.data = self._package(optimized_data)
